#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "menu.h"
#include "obj.h"
#include "mission.h"
#include "simutils.h"
#include "sar.h"

#include "missionio.h"


/* Menu handling utilities and common procedures. */
void SARMenuButtonCBSelectMenu(
	sar_core_struct *core_ptr,
	const char *name
);
int SARMenuObjectGetByID(sar_menu_struct *menu_ptr, int id);
sar_menu_list_struct *SARMenuGetList(
	sar_menu_struct *menu_ptr, int skip, int *list_num
);
sar_menu_spin_struct *SARMenuGetSpin(
	sar_menu_struct *menu_ptr, int skip, int *spin_num
);
sar_menu_list_item_struct *SARMenuListCBGetItem(
	sar_menu_struct *menu_ptr, int skip
);
char *SARMenuGetCurrentMenuName(sar_core_struct *core_ptr);

/* Particular menu callbacks. */
static void SARMenuButtonCBTest(sar_core_struct *core_ptr);
static void SARMenuCBBeginMission(sar_core_struct *core_ptr);
static void SARMenuCBViewLastLog(sar_core_struct *core_ptr);
static void SARMenuCBMissionMap(sar_core_struct *core_ptr);
static void SARMenuCBMissionBrief(sar_core_struct *core_ptr);
static void SARMenuCBBeginFreeFlight(sar_core_struct *core_ptr);
static void SARMenuCBFreeFlightAircraftInfo(sar_core_struct *core_ptr);
static void SARMenuCBMapOp(sar_core_struct *core_ptr, int id_code);

/* Menu widget/objects callback. */
void SARMenuButtonCB(
	void *object, void *client_data,
        int id_code
);
void SARMenuListSelectCB(
        void *object, void *client_data,
	int item_num, void *item_ptr
);
void SARMenuSwitchSelectCB(
        void *object, void *client_data,
        int id_code, Boolean state
);
void SARMenuSpinSelectCB(
	void *object, void *client_data,
	int id_code, char *value
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *      Default spacings (in pixels):
 *      
 *      These values should match and coenciding values in menu.c. 
 */
#define DEF_XMARGIN             5
#define DEF_YMARGIN             5

#define DEF_WIDTH               100
#define DEF_HEIGHT              100


/*
 *	Common callback for menu buttons, sets the current menu on
 *	indicated on the core structure to that matching the given name.
 *
 *	If name is NULL then the current menu on the core structure is set
 *	to -1, indicating no menu is selected and implying simulation is
 *	being displayed.
 */
void SARMenuButtonCBSelectMenu(sar_core_struct *core_ptr, const char *name)
{
	int new_menu_num, old_menu_num;


	if(core_ptr == NULL)
	    return;

        /* Record old (current) menu number. */
        old_menu_num = core_ptr->cur_menu;

	/* Is given menu name NULL? */
	if(name == NULL)
	{
	    /* No menu name given, this implies we are going out of the
	     * menu system and entering simulation.
	     */
	    sar_menu_struct *old_menu = (SARIsMenuAllocated(core_ptr, old_menu_num) ?
                core_ptr->menu[old_menu_num] : NULL
            );

	    /* Unload image on old menu. */
/*	    if(option.prioritize_memory && (old_menu != NULL)) */
	    if(old_menu != NULL)
		SARMenuUnloadBackgroundImage(old_menu);

	    /* Unselect current menu on core structure, indicating that
	     * no menu is selected and that we are not in the menus.
	     */
	    core_ptr->cur_menu = -1;

	    /* Skip redrawing, since redrawing would be done rather
	     * frequently during the simulation.
	     */
/*	    GWPostRedraw(core_ptr->display); */

	    return;
	}

	/* Get menu number to change to by matching the menu who's name
	 * is the same as the given name.
	 */
	new_menu_num = SARMatchMenuByName(core_ptr, name);
	/* Got match and matched menu number is different from the one
	 * currently selected?
	 */
	if((new_menu_num > -1) && (new_menu_num != core_ptr->cur_menu))
	{
            sar_menu_struct *old_menu = (SARIsMenuAllocated(core_ptr, old_menu_num) ?
		core_ptr->menu[old_menu_num] : NULL
	    );
	    sar_menu_struct *new_menu = (SARIsMenuAllocated(core_ptr, new_menu_num) ?
		core_ptr->menu[new_menu_num] : NULL
	    );
	    const char *old_bg_image_path, *new_bg_image_path;


	    /* Mark input as busy while deallocating/allocating resources
	     * during changing menus.
	     */
            GWSetInputBusy(core_ptr->display);


	    /* Get background image paths on both new and old menus, use
	     * NULL if respective menu is NULL.
	     */
	    old_bg_image_path = (const char *)((old_menu != NULL) ?
		old_menu->bg_image_path : NULL
	    );
            new_bg_image_path = (const char *)((new_menu != NULL) ?
                new_menu->bg_image_path : NULL
            );

	    /* Check if background images are the same on both old and
	     * new menus by comparing the background image paths (if
	     * defined).
	     */
	    if(((old_bg_image_path != NULL) &&
                (new_bg_image_path != NULL)) ?
	         ((old_menu->bg_image != NULL) ?
		  !strcmp(old_bg_image_path, new_bg_image_path) : 0)
		: 0
	    )
	    {
		/* Image paths are the same, so unload the image on the
		 * new menu and transfer the image from the old menu to
		 * the new menu. Mark the old menu's image NULL after
		 * transfer so that it dosen't get referenced again.
		 * Note that both new_menu and old_menu are known to be
		 * not NULL from the above check.
		 */
		SARMenuUnloadBackgroundImage(new_menu);
		if(new_menu->bg_image == NULL)
		{
		    new_menu->bg_image = old_menu->bg_image;
		    old_menu->bg_image = NULL;
		}
	    }
	    else
	    {
		/* Image paths are different, so unload background image
		 * on old menu and load new background image on new menu.
		 */
/*		if(option.prioritize_memory && (old_menu != NULL)) */
		if(old_menu != NULL)
		    SARMenuUnloadBackgroundImage(old_menu);

		if(option.menu_backgrounds &&
		   ((new_menu != NULL) ? (new_menu->bg_image == NULL) : 0)
		)
		    SARMenuLoadBackgroundImage(new_menu);
	    }

	    /* Is the new menu valid? */
	    if(new_menu != NULL)
	    {
		/* New menu is valid, so switch to it. */

		/* Reset any progress bars on the new menu. */
                SARMenuSetProgressSimple(
		    core_ptr->display,
		    new_menu,
		    0,
		    0
		);

		/* Update index number of the selected menu on the core
		 * structure.
		 */
	        core_ptr->cur_menu = new_menu_num;
	    }

	    /* Redraw. */
	    GWPostRedraw(core_ptr->display);

	    /* Mark input as ready. */
	    GWSetInputReady(core_ptr->display);
	}
}

/*
 *	Returns the index number of the menu object who's id matches
 *	the given id. Can return -1 on error.
 */
int SARMenuObjectGetByID(sar_menu_struct *menu_ptr, int id)
{
	int i;
	void *ptr;
	sar_menu_button_struct *button_ptr;
	sar_menu_switch_struct *switch_ptr;
	sar_menu_spin_struct *spin_ptr;


	if(menu_ptr == NULL)
	    return(-1);

        for(i = 0; i < menu_ptr->total_objects; i++)
        {
            ptr = menu_ptr->object[i];
            if(ptr == NULL)
                continue;

            switch(*(int *)ptr)
	    {
	      case SAR_MENU_OBJECT_TYPE_BUTTON:
		button_ptr = (sar_menu_button_struct *)ptr;
		if(button_ptr->id_code == id)
		    return(i);
		break;

              case SAR_MENU_OBJECT_TYPE_SWITCH:
                switch_ptr = (sar_menu_switch_struct *)ptr;
                if(switch_ptr->id_code == id)
                    return(i);
                break;

              case SAR_MENU_OBJECT_TYPE_SPIN:
                spin_ptr = (sar_menu_spin_struct *)ptr;
                if(spin_ptr->id_code == id)
                    return(i);
                break;

	    }
	}

	return(-1);
}

/*
 *	Returns the pointer to the list object on the menu,
 *	skipping the indicated number of lists.
 *
 *	Can return NULL for no match.
 */
sar_menu_list_struct *SARMenuGetList(
	sar_menu_struct *menu_ptr, int skip, int *list_num
)     
{
        int i;
        void *ptr;
        sar_menu_list_struct *list_ptr;


        if(menu_ptr == NULL)
            return(NULL);

	if(list_num != NULL)
	    *list_num = -1;
 
        for(i = 0; i < menu_ptr->total_objects; i++)
        {       
            ptr = menu_ptr->object[i];
            if(ptr == NULL)
                continue;

            if((*(int *)ptr) != SAR_MENU_OBJECT_TYPE_LIST)
                continue;

            /* Get pointer to list object. */
            list_ptr = ptr;

            /* Skip this list object? */
            if(skip > 0)
            {
                skip--;
                continue;
            }
	    else
	    {
		if(list_num != NULL)
		    *list_num = i;
                return(list_ptr);
	    }
	}

	return(NULL);
}

/*
 *      Returns the pointer to the spin object on the menu,
 *      skipping the indicated number of spin objects.
 *
 *	Can return NULL for no match.
 */
sar_menu_spin_struct *SARMenuGetSpin(
        sar_menu_struct *menu_ptr, int skip, int *spin_num
)
{
        int i;
        void *ptr;
	sar_menu_spin_struct *spin_ptr;


        if(menu_ptr == NULL)
            return(NULL);

        if(spin_num != NULL)
            *spin_num = -1;

        for(i = 0; i < menu_ptr->total_objects; i++)
        {
            ptr = menu_ptr->object[i];
            if(ptr == NULL)
                continue;
 
            if((*(int *)ptr) != SAR_MENU_OBJECT_TYPE_SPIN)
                continue;

            /* Get pointer to spin object. */
            spin_ptr = ptr;

            /* Skip this spin object? */
            if(skip > 0)
            {
                skip--;
                continue;
            }
            else
            {
                if(spin_num != NULL)
                    *spin_num = i;

                return(spin_ptr);
            }
        }

        return(NULL);
}

/*
 *	Returns the pointer to the selected list item on the list object
 *	on the menu, skipping the indicated number of list objects.
 */
sar_menu_list_item_struct *SARMenuListCBGetItem(
	sar_menu_struct *menu_ptr, int skip
)
{
	int i, sel_num;
	void *ptr;
	sar_menu_list_struct *list_ptr;


	if(menu_ptr == NULL)
	    return(NULL);

	for(i = 0; i < menu_ptr->total_objects; i++)
	{
	    ptr = menu_ptr->object[i];
	    if(ptr == NULL)
		continue;

	    if((*(int *)ptr) != SAR_MENU_OBJECT_TYPE_LIST)
                continue;

	    /* Get pointer to list object. */
	    list_ptr = ptr;

	    /* Skip this list object? */
	    if(skip > 0)
	    {
		skip--;
		continue;
	    }

	    /* Get selected item number. */
	    sel_num = list_ptr->selected_item;

	    /* Return pointer to selected item if valid. */
	    if((sel_num < 0) || (sel_num >= list_ptr->total_items))
		return(NULL);
	    else
		return(list_ptr->item[sel_num]);
	}

	return(NULL);
}

/*
 *	Returns the pointer to the name of the currently selected
 *	menu or NULL on failure.
 */
char *SARMenuGetCurrentMenuName(sar_core_struct *core_ptr)
{
	int cur_menu;
	sar_menu_struct *menu_ptr;


	if(core_ptr == NULL)
	    return(NULL);

	cur_menu = core_ptr->cur_menu;
	if(SARIsMenuAllocated(core_ptr, cur_menu))
	    menu_ptr = core_ptr->menu[cur_menu];
	else
	    return(NULL);

	return(menu_ptr->name);
}


/*
 *	Test callback procedure, used for debugging purposes when
 *	SAR_OP_TEST is given as the operation in SARMenuButtonCB().
 */
static void SARMenuButtonCBTest(sar_core_struct *core_ptr)
{
	int i;
	sar_menu_struct *mission_log_menu;


        if(core_ptr == NULL)
            return;

        /* Get mission log map menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION_LOG_MAP);
        mission_log_menu = ((i >= 0) ? core_ptr->menu[i] : NULL);

        /* Update the map object's map and markings with the information
	 * on the mission log file.
	 */
        SARMissionLoadMissionLogToMenuMap(
            core_ptr, mission_log_menu,
            fname.mission_log
        );

        /* Switch to log map menu. */
        SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MISSION_LOG_MAP);
}

/*
 *	Begin mission callback procedure.
 */
static void SARMenuCBBeginMission(sar_core_struct *core_ptr)
{
        int i;
        sar_menu_struct *menu_ptr;
	const sar_menu_list_item_struct *item_ptr;
	const char *filename;
        sar_menu_list_item_data_struct *li_data_ptr;

  
        if(core_ptr == NULL)
            return;

	/* Begin collecting data from menu objects in order to pass to
	 * begin mission function.
	 */

        /* Get pointer to menu Mission. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

        /* Get pointer to selected item on first list object. */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
        li_data_ptr = ((item_ptr == NULL) ? NULL : item_ptr->client_data);
        filename = ((li_data_ptr == NULL) ? NULL : li_data_ptr->filename);
        free(core_ptr->cur_mission_file);
        core_ptr->cur_mission_file = StringCopyAlloc(filename);


	/* Switch to mission simulation and leave menu system. */
	SARSimBeginMission(
	    core_ptr,
	    core_ptr->cur_mission_file
	);
}

/*
 *	Switch to last log map procedure.
 */
static void SARMenuCBViewLastLog(sar_core_struct *core_ptr)
{
        int i;
        sar_menu_struct *mission_log_menu;


        if(core_ptr == NULL)
            return;

        /* Get mission log map menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION_LOG_MAP);
        mission_log_menu = ((i >= 0) ? core_ptr->menu[i] : NULL);

        /* Update the map object's map and markings with the information
         * on the mission log file.
         */
        SARMissionLoadMissionLogToMenuMap(
            core_ptr, mission_log_menu,
            fname.mission_log
        );

        /* Switch to log map menu. */
        SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MISSION_LOG_MAP);
}

/*
 *	Switch to mission map procedure.
 */
static void SARMenuCBMissionMap(sar_core_struct *core_ptr)
{
	int i;
        sar_menu_struct *menu_ptr;
        sar_menu_list_item_struct *item_ptr;
        char *filename;
        sar_menu_list_item_data_struct *li_data_ptr;


        if(core_ptr == NULL)
            return;


        /* Get pointer to mission menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

        /* Get pointer to selected item on first list object. */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
        li_data_ptr = ((item_ptr == NULL) ? NULL : item_ptr->client_data);
        filename = ((li_data_ptr == NULL) ? NULL : li_data_ptr->filename);


        /* Get pointer to mission map menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION_MAP);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

	/* Load scene's map and markings to the map object from the given
	 * scene file.
	 */
	SARMissionLoadSceneToMenuMap(
	    core_ptr, menu_ptr, filename
	);

	/* Switch to mission map menu. */
        SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MISSION_MAP);
}

/*
 *	Switch to mission briefing procedure.
 */
static void SARMenuCBMissionBrief(sar_core_struct *core_ptr)
{
        int i;
        sar_menu_struct *menu_ptr;
        sar_menu_list_item_struct *item_ptr;
        char *filename;
        sar_menu_list_item_data_struct *li_data_ptr;
	char *mission_desc;


        if(core_ptr == NULL)
            return;


        /* Get pointer to mission menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

        /* Get pointer to selected item on first list object. */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
        li_data_ptr = ((item_ptr == NULL) ? NULL : item_ptr->client_data);
        filename = ((li_data_ptr == NULL) ? NULL : li_data_ptr->filename);


        /* Get pointer to mission brief menu. */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_MISSION_BRIEF);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);


	/* Get mission description (briefing). */
	mission_desc = SARMissionLoadDescription(filename);

        /* Set text for message box on briefing menu. */
        SARMenuSetMessageBoxSimple(
            core_ptr->display,
            menu_ptr,
            mission_desc,
            0
        );

        /* Switch to mission brief menu. */
        SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MISSION_BRIEF);
}

/*
 *	Begin free flight procedure.
 */
static void SARMenuCBBeginFreeFlight(sar_core_struct *core_ptr)
{
	int i;
	sar_menu_struct *menu_ptr;
	sar_menu_list_item_struct *item_ptr;
	const char *filename;
	sar_menu_list_item_data_struct *li_data_ptr;
	int spin_num, spin_sel_item;
	sar_menu_switch_struct *switch_ptr;
	sar_menu_spin_struct *spin_ptr;
	sar_position_struct *start_pos = NULL;
	sar_direction_struct *start_dir = NULL;
	const char *weather_preset_name = NULL;
	Boolean free_flight_system_time = False;


	if(core_ptr == NULL)
	    return;

	/* Begin collecting information from menu objects in order to
	 * pass to the begin free flight simulation function.
	 */

	/* Get pointer to menu Free Flight: Select Scenery */
	i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_FREE_FLIGHT);
	menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

	/* Scenery (first list object). */
	item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
	li_data_ptr = (item_ptr != NULL) ? item_ptr->client_data : NULL;
 	filename = (const char *)((li_data_ptr != NULL) ?
	    li_data_ptr->filename : NULL
	);
	free(core_ptr->cur_scene_file);
	core_ptr->cur_scene_file = StringCopyAlloc(filename);

        /* Starting location (second list object). */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 1);
        li_data_ptr = (item_ptr != NULL) ? item_ptr->client_data : NULL;
 	if(li_data_ptr != NULL)
	{
            start_pos = &li_data_ptr->pos;
	    start_dir = &li_data_ptr->dir;
	}


        /* Get pointer to menu Free Flight: Select Aircraft */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

        /* Aircraft (first list object). */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
        li_data_ptr = (item_ptr != NULL) ? item_ptr->client_data : NULL;
        filename = (const char *)((li_data_ptr != NULL) ?
	    li_data_ptr->filename : NULL
	);
        free(core_ptr->cur_player_model_file);
        core_ptr->cur_player_model_file = StringCopyAlloc(filename);


	/* Get pointer to menu Free Flight: Weather */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_WEATHER);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

	/* Weather preset name (first spin object). */
	i = SARMenuObjectGetByID(
	    menu_ptr, SAR_OP_MENU_FREE_FLIGHT_WEATHER_CONDITION
	);
	spin_num = i;
	spin_ptr = (sar_menu_spin_struct *)((i >= 0) ?
	    menu_ptr->object[i] : NULL
	);
	weather_preset_name = (const char *)SARMenuSpinGetCurrentValue(
	    menu_ptr, spin_num, &spin_sel_item
	);

	/* Free flight system time (first switch). */
        i = SARMenuObjectGetByID(
            menu_ptr, SAR_OP_MENU_FREE_FLIGHT_SYSTEM_TIME
        );
        switch_ptr = (sar_menu_switch_struct *)((i >= 0) ?
            menu_ptr->object[i] : NULL
        );
	free_flight_system_time = switch_ptr->state;


	/* Switch to free flight simulation and leave menu system. */
	SARSimBeginFreeFlight(
	    core_ptr,
	    core_ptr->cur_scene_file,
	    core_ptr->cur_player_model_file,
	    start_pos, start_dir,
	    weather_preset_name, free_flight_system_time
	);
}

/*
 *	Switch to aircraft information menu procedure.
 */
static void SARMenuCBFreeFlightAircraftInfo(sar_core_struct *core_ptr)
{
        int i, ov_num;
        sar_menu_struct *menu_ptr;
        sar_menu_list_item_struct *item_ptr;
        const char *filename;
        sar_menu_list_item_data_struct *li_data_ptr;
	sar_menu_objview_struct *ov_ptr;


        if(core_ptr == NULL)
            return;

        /* Get pointer to menu Free Flight: Select Aircraft */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

        /* Aircraft (first list object). */
        item_ptr = SARMenuListCBGetItem(menu_ptr, 0);
        li_data_ptr = (item_ptr != NULL) ? item_ptr->client_data : NULL;
        filename = (const char *)((li_data_ptr != NULL) ?
            li_data_ptr->filename : NULL
        );


        /* Get pointer to menu Free Flight: Aircraft information */
        i = SARMatchMenuByName(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT_INFO);
        menu_ptr = ((i < 0) ? NULL : core_ptr->menu[i]);

	/* Get pointer to aircraft information 3d object viewer object
	 * on the menu.
	 */
	ov_num = -1;
	ov_ptr = NULL;
	if(menu_ptr != NULL)
	{
	    ov_num = 0;

	    if((ov_num >= 0) && (ov_num < menu_ptr->total_objects))
	    {
		void *p = menu_ptr->object[ov_num];
		if((p != NULL) ?
		    ((*(int *)p) == SAR_MENU_OBJECT_TYPE_OBJVIEW) : 0
		)
		    ov_ptr = (sar_menu_objview_struct *)p;
	    }
	}
	if(ov_ptr != NULL)
	{
            /* Mark input as busy while loading new aircraft information,
	     * note that the input will be set busy again when we
	     * switch menus but that is okay.
	     */
            GWSetInputBusy(core_ptr->display);

	    /* Set up aircraft information 3d object view to display
	     * new aircraft.
	     */
	    SARMenuObjViewLoad(
		core_ptr->display, menu_ptr,
		ov_ptr, ov_num, filename
	    );
	    SARMenuObjViewRebufferImage(
		core_ptr->display, menu_ptr,
		ov_ptr, ov_num
	    );
	}

	/* Switch to aircraft information menu. */
	SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT_INFO);
}


/*
 *	Map menu button callback.
 */
static void SARMenuCBMapOp(sar_core_struct *core_ptr, int id_code)
{
	gw_display_struct *display;
	int i, ptype, map_num = -1;
	int width, height, w, h;
	void *p;
        sar_menu_struct *m;
	sar_menu_map_struct *map_ptr = NULL;

	double mtop_coeff;
	int	scroll_inc = 20;	/* Scroll increment in pixels. */
	int	tex_w, tex_h,	/* All these units in pixels. */
		scroll_x_min, scroll_x_max,
		scroll_y_min, scroll_y_max;


	if(core_ptr == NULL)
	    return;

	display = core_ptr->display;
	if(display == NULL)
	    return;

        /* Get current menu (assuming it has a map object on it. */
	m = SARGetCurrentMenuPtr(core_ptr);
	if(m == NULL)
	    return;

        /* Get first map object on the menu. */
        for(i = 0; i < m->total_objects; i++)
        {
            p = m->object[i];
            if(p == NULL)
                continue;
            else
                ptype = (*(int *)p);

            if(ptype == SAR_MENU_OBJECT_TYPE_MAP)
            {
                map_num = i;
                map_ptr = (sar_menu_map_struct *)p;
                break;
            }
        }
        if(map_ptr == NULL)
            return;

/* Calculates scroll bounds by first fetching latest mtop_coeff
 * value and updates variables tex_w, tex_h, scroll_x_min, scroll_y_min,
 * scroll_x_max, and scroll_y_max.
 */
#define GET_SCROLL_BOUNDS	\
{ \
 mtop_coeff = map_ptr->m_to_pixels_coeff; \
 \
 /* Get texture size in pixels. */ \
 tex_w = (int)(map_ptr->bg_tex_width * mtop_coeff); \
 tex_h = (int)(map_ptr->bg_tex_height * mtop_coeff); \
 \
 /* Calculate scroll bounds in pixels. */ \
 scroll_x_min = (w / 2) - (tex_w / 2); \
 scroll_x_max = (tex_w / 2) - (w / 2); \
 \
 scroll_y_min = (h / 2) - (tex_h / 2); \
 scroll_y_max = (tex_h / 2) - (h / 2); \
}

	/* Get bounds of display. */
	GWContextGet(
	    display, 0,
	    NULL, NULL,
	    NULL, NULL,
	    &width, &height
	);

	/* Get bounds of map object. */
        w = (int)(map_ptr->width * width);
        if(w <= 0)
            w = (int)(width - (2 * DEF_XMARGIN));
        if(w <= 0)
            w = DEF_XMARGIN;

        if(map_ptr->height > 0)
            h = (int)(map_ptr->height * height);
        else
            h = DEF_HEIGHT;
        if(h <= 0)
            h = DEF_YMARGIN;


	/* Calculate scroll bounds. */
	GET_SCROLL_BOUNDS

	/* Handle by id code. */
	switch(id_code)
	{
	  case SAR_OP_MISSION_MAP_LEFT: case SAR_OP_MISSION_LOG_MAP_LEFT:
/*
	    map_ptr->scroll_x = CLIP(
		map_ptr->scroll_x + scroll_inc,
		scroll_x_min, scroll_x_max
	    );
 */
	    map_ptr->scroll_x = map_ptr->scroll_x + scroll_inc;
	    break;

          case SAR_OP_MISSION_MAP_RIGHT: case SAR_OP_MISSION_LOG_MAP_RIGHT:
/*
            map_ptr->scroll_x = CLIP(
                map_ptr->scroll_x - scroll_inc,
                scroll_x_min, scroll_x_max
            );
 */
	    map_ptr->scroll_x = map_ptr->scroll_x - scroll_inc;
            break;

          case SAR_OP_MISSION_MAP_UP: case SAR_OP_MISSION_LOG_MAP_UP:
/*
            map_ptr->scroll_y = CLIP(
                map_ptr->scroll_y - scroll_inc,
                scroll_y_min, scroll_y_max
            );
 */
	    map_ptr->scroll_y = map_ptr->scroll_y - scroll_inc;
            break;

          case SAR_OP_MISSION_MAP_DOWN: case SAR_OP_MISSION_LOG_MAP_DOWN:
/*
            map_ptr->scroll_y = CLIP(
                map_ptr->scroll_y + scroll_inc,
                scroll_y_min, scroll_y_max
            );
 */
	    map_ptr->scroll_y = map_ptr->scroll_y + scroll_inc;
            break;

          case SAR_OP_MISSION_MAP_ZOOM_IN: case SAR_OP_MISSION_LOG_MAP_ZOOM_IN:
	    /* This portion of code came from menumap.c's drag zoom
	     * event handling.
	     */
	    {
                    double scroll_x_coeff, scroll_y_coeff;
                    double prev_m_to_pixels_coeff =
                        map_ptr->m_to_pixels_coeff;

                    /* Calculate new zoom. */
                    map_ptr->m_to_pixels_coeff = CLIP(
                        map_ptr->m_to_pixels_coeff +
                            (scroll_inc *
                            map_ptr->m_to_pixels_coeff / 100),
                        0.002,          /* Zoomed out max. */
                        0.050           /* Zoomed in max. */
                    );

                    /* Need to calculate last scroll position coeff
                     * relative to the previous zoom dimension, then
                     * set the new scroll position by multiplying
                     * the scroll position coeff just calculated by
                     * the new zoom dimension.
                     */
                    if(map_ptr->bg_tex_width > 0)
                        scroll_x_coeff = (double)map_ptr->scroll_x /
                            (map_ptr->bg_tex_width *
                                prev_m_to_pixels_coeff);
                    else
                        scroll_x_coeff = 0.0;
                    map_ptr->scroll_x = (int)(map_ptr->bg_tex_width *
                        map_ptr->m_to_pixels_coeff * scroll_x_coeff);

                    if(map_ptr->bg_tex_height > 0)
                        scroll_y_coeff = -(double)map_ptr->scroll_y /
                            (map_ptr->bg_tex_height *
                                prev_m_to_pixels_coeff);
                    else
                        scroll_y_coeff = 0.0;

                    map_ptr->scroll_y = -(int)(map_ptr->bg_tex_height *
                        map_ptr->m_to_pixels_coeff * scroll_y_coeff);

                    /* Need to calculate scroll bounds again. */
                    GET_SCROLL_BOUNDS

                    /* Reclip scroll bounds. */
/* Its safe to scroll anywhere, plus we can see offmap markings.
                    map_ptr->scroll_x = CLIP(
                        map_ptr->scroll_x,
                        scroll_x_min, scroll_x_max
                    );
                    map_ptr->scroll_y = CLIP(
                        map_ptr->scroll_y,
                        scroll_y_min, scroll_y_max
                    );
 */
	    }
	    break;

          case SAR_OP_MISSION_MAP_ZOOM_OUT: case SAR_OP_MISSION_LOG_MAP_ZOOM_OUT:
            /* This portion of code came from menumap.c's drag zoom  
             * event handling.
             */
            {
                    double scroll_x_coeff, scroll_y_coeff;
                    double prev_m_to_pixels_coeff =
                        map_ptr->m_to_pixels_coeff;

                    /* Calculate new zoom. */
                    map_ptr->m_to_pixels_coeff = CLIP(
                        map_ptr->m_to_pixels_coeff -
                            (scroll_inc *
                            map_ptr->m_to_pixels_coeff / 100),
                        0.002,          /* Zoomed out max. */
                        0.050           /* Zoomed in max. */
                    );

                    /* Need to calculate last scroll position coeff
                     * relative to the previous zoom dimension, then
                     * set the new scroll position by multiplying
                     * the scroll position coeff just calculated by
                     * the new zoom dimension.
                     */
                    if(map_ptr->bg_tex_width > 0)
                        scroll_x_coeff = (double)map_ptr->scroll_x /
                            (map_ptr->bg_tex_width *
                                prev_m_to_pixels_coeff);
                    else
                        scroll_x_coeff = 0.0;
                    map_ptr->scroll_x = (int)(map_ptr->bg_tex_width *
                        map_ptr->m_to_pixels_coeff * scroll_x_coeff);

                    if(map_ptr->bg_tex_height > 0)
                        scroll_y_coeff = -(double)map_ptr->scroll_y /
                            (map_ptr->bg_tex_height * 
                                prev_m_to_pixels_coeff);
                    else
                        scroll_y_coeff = 0.0;

                    map_ptr->scroll_y = -(int)(map_ptr->bg_tex_height *
                        map_ptr->m_to_pixels_coeff * scroll_y_coeff);

                    /* Need to calculate scroll bounds again. */   
                    GET_SCROLL_BOUNDS

                    /* Reclip scroll bounds. */
/* Its safe to scroll anywhere, plus we can see offmap markings.
                    map_ptr->scroll_x = CLIP(
                        map_ptr->scroll_x,
                        scroll_x_min, scroll_x_max
                    );
                    map_ptr->scroll_y = CLIP(
                        map_ptr->scroll_y,
                        scroll_y_min, scroll_y_max
                    );
 */
            }
            break;

	  case SAR_OP_MISSION_LOG_MAP_EVENT_NEXT:
	    SARMenuMapMarkingSelectNext(map_ptr, True);
	    break;

	  case SAR_OP_MISSION_LOG_MAP_EVENT_PREV:
            SARMenuMapMarkingSelectPrev(map_ptr, True);
	    break;
	}

	/* Redraw map object. */
	SARMenuDrawObject(
	    display, m, map_num
	);
	GWSwapBuffer(display);

#undef GET_SCROLL_BOUNDS
}


/*
 *	Menu button callback.
 */
void SARMenuButtonCB(
	void *object, void *client_data, int id_code
)
{
	sar_core_struct *core_ptr = (sar_core_struct *)client_data;


	if((core_ptr == NULL) || (object == NULL))
	    return;

	switch(id_code)
	{
	  case SAR_OP_MENU_TO_MAIN:
	    SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MAIN);
	    break;

	  case SAR_OP_MENU_EXIT:
            runlevel = 1;
            break;

	  case SAR_OP_MENU_TO_FREE_FLIGHT:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_FREE_FLIGHT);
	    break;

	  case SAR_OP_MENU_TO_FREE_FLIGHT_AIRCRAFT:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT);
            break;

          case SAR_OP_MENU_TO_FREE_FLIGHT_WEATHER:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_FREE_FLIGHT_WEATHER);
            break;

	  case SAR_OP_MENU_TO_FREE_FLIGHT_AIRCRAFT_INFO:
	    SARMenuCBFreeFlightAircraftInfo(core_ptr);
            break;

	  case SAR_OP_MENU_TO_FREE_FLIGHT_BEGIN:
            SARMenuCBBeginFreeFlight(core_ptr);
            break;

	  case SAR_OP_MENU_TO_MISSION:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_MISSION);
	    break;

	  case SAR_OP_MENU_TO_MISSION_LOG_MAP:
	    SARMenuCBViewLastLog(core_ptr);
	    break;

          case SAR_OP_MENU_TO_MISSION_MAP:
	    SARMenuCBMissionMap(core_ptr);
            break;

          case SAR_OP_MENU_TO_MISSION_BRIEF:
	    SARMenuCBMissionBrief(core_ptr);
            break;

	  case SAR_OP_MENU_TO_MISSION_BEGIN:
	    SARMenuCBBeginMission(core_ptr);
	    break;

	  case SAR_OP_MISSION_MAP_LEFT:
	  case SAR_OP_MISSION_MAP_RIGHT:
	  case SAR_OP_MISSION_MAP_UP:
	  case SAR_OP_MISSION_MAP_DOWN:
	  case SAR_OP_MISSION_MAP_ZOOM_IN:
          case SAR_OP_MISSION_MAP_ZOOM_OUT:
          case SAR_OP_MISSION_LOG_MAP_LEFT:
          case SAR_OP_MISSION_LOG_MAP_RIGHT:
          case SAR_OP_MISSION_LOG_MAP_UP:
          case SAR_OP_MISSION_LOG_MAP_DOWN:
          case SAR_OP_MISSION_LOG_MAP_ZOOM_IN:
          case SAR_OP_MISSION_LOG_MAP_ZOOM_OUT:
	  case SAR_OP_MISSION_LOG_MAP_EVENT_NEXT:
	  case SAR_OP_MISSION_LOG_MAP_EVENT_PREV:
            SARMenuCBMapOp(core_ptr, id_code);   
            break;

	  case SAR_OP_MENU_TO_CAMPAIGN:
/*
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_CAMPAIGN);
 */
            GWOutputMessage(
                core_ptr->display, GWOutputMessageTypeWarning,
                "Campaign Support Not Finished",
 "Sorry, campaign support is still under construction.",
"The coding work for campaign support is still under construction\n\
and not yet playable, please choose to either fly a mission or\n\
free flight."
            );

	    break;

	  case SAR_OP_MENU_TO_OPTIONS:
	    SARFetchOptionsToMenus(core_ptr);
	    SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_OPTIONS);
	    break;

          case SAR_OP_MENU_TO_OPTIONS_SIMULATION:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_OPTIONS_SIMULATION);
            break;

          case SAR_OP_MENU_TO_OPTIONS_CONTROLLER:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_OPTIONS_CONTROLLER);
            break;

	  case SAR_OP_MENU_TO_OPTIONS_CONTROLLER_JS_BTN:
	    SARMenuButtonCBSelectMenu(
		core_ptr, SAR_MENU_NAME_OPTIONS_CONTROLLER_JS_BTN
	    );
            break;

          case SAR_OP_MENU_TO_OPTIONS_GRAPHICS:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_OPTIONS_GRAPHICS);
            break;

          case SAR_OP_MENU_TO_OPTIONS_SOUNDS:
            SARMenuButtonCBSelectMenu(core_ptr, SAR_MENU_NAME_OPTIONS_SOUND);
            break;


case SAR_OP_TEST:
 SARMenuButtonCBTest(core_ptr);
 break;




	  default:
	    fprintf(stderr,
 "SARMenuButtonCB(): Internal error: Unsupported menu operation code `%i'.\n",
		id_code
	    );
	    break;
	}
}


/*
 *	Menu standard list object callback.
 */
void SARMenuListSelectCB(
	void *object, void *client_data,
	int item_num, void *item_ptr
)
{
	char *cur_menu_name;
	sar_core_struct *core_ptr;
	sar_menu_struct *cur_menu_ptr;
	sar_menu_list_struct *list_ptr;
	sar_menu_list_item_data_struct *item_data_ptr;


	list_ptr = object;
	if(list_ptr == NULL)
	    return;

	item_data_ptr = client_data;
	if(item_data_ptr == NULL)
	    return;

	core_ptr = item_data_ptr->core_ptr;
	if(core_ptr == NULL)
	    return;


	/* Get current selected menu and name. */
	cur_menu_ptr = SARGetCurrentMenuPtr(core_ptr);
	if(cur_menu_ptr == NULL)
	    return;

	cur_menu_name = cur_menu_ptr->name;
	if(cur_menu_name == NULL)
	    return;



	/* Mission: Select Mission */
	if(!strcasecmp(cur_menu_name, SAR_MENU_NAME_MISSION))
	{
	    option.last_selected_mission = item_num;
	}
	/* Free Flight: Select Scenery */
        else if(!strcasecmp(cur_menu_name, SAR_MENU_NAME_FREE_FLIGHT))
        {
	    if(item_data_ptr->filename != NULL)
	    {
		int loc_list_num;
		sar_menu_list_struct *loc_list_ptr;

		/* Specified filename implies this is the item data from
		 * the scenery list (not the locations list on the same
		 * menu).
		 */

		/* Get pointer to second list (skip 1). */
		loc_list_ptr = SARMenuGetList(cur_menu_ptr, 1, &loc_list_num);
		if(loc_list_ptr != NULL)
		{
		    /* Load registered locations from scene, any
		     * existing locations on the list will be deleted.
		     */
		    SARSceneLoadLocationsToList(
		        core_ptr,
		        cur_menu_ptr, loc_list_ptr, loc_list_num,
		        item_data_ptr->filename
		    );

		    /* Select first item on locations list. */
		    if(loc_list_ptr->total_items > 0)
		        loc_list_ptr->selected_item = 0;
		}

		SARMenuDrawAll(core_ptr->display, cur_menu_ptr);
		GWPostRedraw(core_ptr->display);
	    }

            option.last_selected_ffscene = item_num;
	}
        /* Free Flight: Select Aircraft */
        else if(!strcasecmp(cur_menu_name, SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT))
        {

            option.last_selected_ffaircraft = item_num;
	}
}

/*
 *      Menu standard switch object callback.
 */
void SARMenuSwitchSelectCB(
        void *object, void *client_data,
        int id_code, Boolean state
)
{
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        sar_menu_switch_struct *switch_ptr = (sar_menu_switch_struct *)object;


        if((core_ptr == NULL) || (switch_ptr == NULL))
            return;

        /* Handle by menu object id code. */
        switch(id_code)
        {
          case SAR_OP_MENU_FREE_FLIGHT_SYSTEM_TIME:
            option.system_time_free_flight = switch_ptr->state;
            break;

        }
}

/*
 *	Menu standard spin object callback.
 */
void SARMenuSpinSelectCB(
	void *object, void *client_data,
	int id_code, char *value
)
{
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        sar_menu_spin_struct *spin_ptr = (sar_menu_spin_struct *)object;


        if((core_ptr == NULL) || (spin_ptr == NULL))
            return;

	/* Handle by menu object id code. */
	switch(id_code)
	{
	  case SAR_OP_MENU_FREE_FLIGHT_WEATHER_CONDITION:
            option.last_selected_ffweather = spin_ptr->cur_value;
	    break;

        }
}
