#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

#ifdef HAVE_Y2
# include <Y2/Y.h>	/* We get Y_H defined by #including this. */
# include <Y2/Ylib.h>
#endif	/* HAVE_Y2 */

#include "../include/string.h"
#include "../include/disk.h"

#include "sound.h"
#include "sar.h"


#ifdef Y_H
/* Default values for Y. */
# ifndef DEF_Y_CONNECT_ARG
#  define DEF_Y_CONNECT_ARG		"127.0.0.1:9433"
# endif
# ifndef DEF_Y_AUDIO_MODE_NAME
#  define DEF_Y_AUDIO_MODE_NAME		"Default"
# endif
#endif	/* Y_H */


snd_recorder_struct *SoundInit(
        int type,
        const char *connect_arg,
        const char *start_arg
);
int SoundManageEvents(snd_recorder_struct *recorder);
void SoundShutdown(snd_recorder_struct *recorder);

int SoundChangeMode(
	snd_recorder_struct *recorder, const char *mode_name
);

snd_play_struct *SoundStartPlay(
	snd_recorder_struct *recorder,
	const char *object,
	double volume_left, double volume_right,
	int sample_rate,
	snd_flags_t options
);
void SoundStartPlayVoid(
	snd_recorder_struct *recorder,
	const char *object,
	double volume_left, double volume_right,
	int sample_rate,
	snd_flags_t options
);

void SoundPlayMute(
	snd_recorder_struct *recorder,
	snd_play_struct *snd_play,
	int mute
);
void SoundChangePlayVolume(
	snd_recorder_struct *recorder,
	snd_play_struct *snd_play,
	double volume_left, double volume_right
);
void SoundChangePlaySampleRate(
	snd_recorder_struct *recorder,
	snd_play_struct *snd_play,
	int sample_rate
);

void SoundStopPlay(
	snd_recorder_struct *recorder,
	snd_play_struct *snd_play
);

int SoundMusicStartPlay(
        snd_recorder_struct *recorder,
        const char *object,     /* Full path to object. */
        int repeats             /* Number of repeats, -1 for infinate. */
);
int SoundMusicIsPlaying(snd_recorder_struct *recorder);
void SoundMusicStopPlay(snd_recorder_struct *recorder);


/*
 *	Initializes sound server and returns a newly allocated
 *	recorder structure.
 */
snd_recorder_struct *SoundInit(
	int type,
	const char *connect_arg,
	const char *start_arg
)
{
	snd_recorder_struct *recorder = NULL;


	/* Allocate recorder structure. */
	recorder = (snd_recorder_struct *)calloc(
	    1, sizeof(snd_recorder_struct)
	);
	if(recorder == NULL)
	    return(NULL);

	recorder->type = type;
	recorder->con = NULL;
	recorder->sample_rate = 0;
	recorder->sample_size = 0;
	recorder->channels = 0;
	recorder->bytes_per_cycle = 0;
	recorder->background_music_sndobj = NULL;

	/* Initialize by sound server type. */
	switch(type)
	{
	  case SNDSERV_TYPE_Y:
#ifdef Y_H
	    if(connect_arg == NULL)
		connect_arg = (const char *)getenv("RECORDER");
	    if(connect_arg == NULL)
		connect_arg = DEF_Y_CONNECT_ARG;

	    /* Connect to sound server. */
	    recorder->con = (void *)YOpenConnection(
		start_arg,
		connect_arg
	    );
	    if(recorder->con == NULL)
	    {
		free(recorder);
		return(NULL);
	    }

	    /* Record sound server type. */
	    recorder->type = type;

	    if(option.runtime_debug)
		printf(
"SoundInit(): Initialized sound server at \"%s\".\n",
		    connect_arg
		);
#else
	    fprintf(
		stderr,
		"SoundInit(): Y2 support not compiled.\n"
	    );
            free(recorder);   
            return(NULL);
#endif	/* Y_H */
	    break;

	  default:
	    fprintf(
		stderr,
		"SoundInit(): Unsupported sound server type `%i'.\n",
		type
	    );
	    free(recorder);
	    return(NULL);
	    break;
	}

	return(recorder);
}


/*
 *	Manages sound server events.
 *
 *	Returns the number of events recieved or -1 on error.
 *
 *	If -1 is returned then recorder has become invalid and should
 *	not be referenced again. This usually means the sound server
 *	has shut down, the connection to the sound server was broken, or 
 *	some other fatal error occured.
 */
int SoundManageEvents(snd_recorder_struct *recorder)
{
	int events_handled = 0;


	if(recorder == NULL)
	    return(-1);

	switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H

/* Macro to return the YID of a given pointer to a snd_play_struct. */
#define GET_SND_PLAY_YID(p)		\
 (((p) != NULL) ? (YID)((snd_play_struct *)(p)->data) : YIDNULL)

/* Macro to set the given snd_play_struct pointer's member data to
 * YIDNULL.
 */
#define SET_SND_PLAY_YID_NULL(p)	\
 { if((p) != NULL) (snd_play_struct *)(p)->data = (void *)YIDNULL; }

            if(recorder->con != NULL)
            {
		YEvent event;

		/* Begin queued Y events. */
                while(YGetNextEvent(
		    (YConnection *)recorder->con,
                    &event,
                    False		/* Do not block. */
                ) > 0)
                {
                    events_handled++;

                    switch(event.type)
                    {
		      case YDisconnect: case YShutdown:
			/* Y server has disconnected or shutdown, the
			 * connection structure needs to be closed,
			 * deallocated, and set to NULL.
			 */
			YCloseConnection(
			    (YConnection *)recorder->con, False
			);
			recorder->con = NULL;

if(event.type == YShutdown)
 fprintf(
  stderr,
  "Lost connection to Y server: Y server has shut down.\n"
 );
else
 fprintf(
  stderr,
  "Lost connection to Y server: Y server has disconnected us.\n"
 );

			/* Deallocate recorder structure and return
			 * indicating that the recorder should not be 
			 * referenced again.
			 */
			free(recorder);
			return(-1);
			break;

                      /* Sound object has stopped playing. */
                      case YSoundObjectKill:
			if(event.kill.yid != YIDNULL)
			{
			    YID yid = event.kill.yid;

			    /* Begin checking which sound object this is. */
			    if(yid == GET_SND_PLAY_YID(recorder->background_music_sndobj))
			    {
				SET_SND_PLAY_YID_NULL(recorder->background_music_sndobj);
				SoundStopPlay(
				    recorder,
				    recorder->background_music_sndobj
				);
				recorder->background_music_sndobj = NULL;
                            }
                        }
                        break; 

                      case YMixerChannel:
                        break;
                    }
		}	/* Handle queued Y events. */
	    }
#undef SET_SND_PLAY_YID_NULL
#undef GET_SND_PLAY_YID
#endif  /* Y_H */
            break;

	  default:
	    free(recorder);
	    return(-1);
            break;
        }

	return(events_handled);
}

/*
 *      Shuts down the sound server, all resources on the given recorder
 *	structure will be deallocated and the recorder structure itself
 *	will be deallocated. The recorder should not be referenced again
 *	after this call.
 */
void SoundShutdown(snd_recorder_struct *recorder)
{
	if(recorder == NULL)
	    return;

	/* Stop related resources first. */
	SoundMusicStopPlay(recorder);

	/* Shut down sound server by type. */
        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if(recorder->con != NULL)
            {
                /* Kill previous background music play. */
/*
                if(sound.bkg_playid != NULL)
                {
                    YDestroyPlaySoundObject(
                        (YConnection *)sound.con_data,
                        (YID)sound.bkg_playid
                    );
                    sound.bkg_playid = NULL;
                }
 */
                /* Close connection to server. */
                YCloseConnection(
                    (YConnection *)recorder->con,
                    False
                );

		if(option.runtime_debug)
		    printf(
"SoundShutdown(): Disconnected from sound server.\n"
		    );

            }
#endif  /* Y_H */
            break;

	}

	/* Deallocate recorder structure. */
	free(recorder);

	if(option.runtime_debug)
	    printf(
"SoundShutdown(): Shutdown done.\n"
	    );
}



/*
 *	Changes audio mode to the one specified by mode_name.
 *
 *	Returns 0 on success or -1 on error.
 */
int SoundChangeMode(
        snd_recorder_struct *recorder, const char *mode_name
)
{
        if(recorder == NULL)
            return(-1);

        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
	    if(recorder->con != NULL)
	    {
		YAudioModeValuesStruct **ptr, *yaudio_mode;
                int i, total_yaudio_modes;

		int sample_rate = 0;
		int sample_size = 0;
		int channels = 0;
		int fragment_size_bytes = 0;


                if(mode_name == NULL)
                    mode_name = DEF_Y_AUDIO_MODE_NAME;

                /* Get listing of available Audio mode. */
                ptr = YGetAudioModes(
                    (YConnection *)recorder->con,
                    &total_yaudio_modes
                );

                /* Check if specified Audio mode is in list. */
                for(i = 0; i < total_yaudio_modes; i++)  
                {
		    yaudio_mode = ptr[i];
                    if(yaudio_mode == NULL)
                        continue;

		    if(yaudio_mode->name == NULL)
			continue;

                    /* Audio mode names match? */
                    if(!strcasecmp(yaudio_mode->name, mode_name))
		    {
			sample_rate = yaudio_mode->sample_rate;
			sample_size = yaudio_mode->sample_size;
			channels = yaudio_mode->sample_rate;
			fragment_size_bytes = yaudio_mode->fragment_size_bytes;
                        break;
		    }
                }
                /* Free Audio modes listing. */
                YFreeAudioModesList(ptr, total_yaudio_modes);

	        /* Audio mode name in list? */
	        if(i >= total_yaudio_modes)
		{
		    /* Could not match audio mode name from server's list
		     * of audio mode names. So we need to return -1 here.
		     */
		    return(-1);
		}

	        /* Change Audio mode. */
                if(YChangeAudioModePreset(
                    (YConnection *)recorder->con,
                    mode_name		/* Audio mode name. */
                ) < 0)
		    return(-1);

		/* Record new Audio values on our recorder structure. */
		recorder->sample_rate = sample_rate;
		recorder->sample_size = sample_size;
		recorder->channels = channels;
		recorder->bytes_per_cycle = fragment_size_bytes;
	    }
#endif	/* Y_H */
            break;

        }

	return(0);
}


/*
 *	Plays a sound object by given object path. Returns a dynamically
 *	allocated structure containing the play information or NULL on failure.
 */
snd_play_struct *SoundStartPlay(
        snd_recorder_struct *recorder,
        const char *object,	/* Full path to sound object. */
        double volume_left,     /* Volume, from 0.0 to 1.0. */
        double volume_right,
        int sample_rate,        /* Applied sample rate, can be 0. */
        snd_flags_t options     /* Any of SND_PLAY_OPTION_*. */
)
{
	snd_play_struct *snd_play = NULL;


	if((recorder == NULL) || (object == NULL))
	    return(snd_play);

        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
	    if(recorder->con != NULL)
	    {
		YID yid;
		YEventSoundPlay value;


		/* Set up sound play values. */
		value.flags = YPlayValuesFlagPosition |
                              YPlayValuesFlagTotalRepeats |
                              YPlayValuesFlagVolume |
                              YPlayValuesFlagSampleRate;
		value.position = 0;
		value.total_repeats = ((options & SND_PLAY_OPTION_REPEATING) ?
		    -1 : 1
		);
		value.left_volume = ((options & SND_PLAY_OPTION_MUTE) ?
		    0 : volume_left
		);
		value.right_volume = ((options & SND_PLAY_OPTION_MUTE) ?
		    0 : volume_right
		);
		value.sample_rate = sample_rate;

		/* Start playing sound object. */
                yid = YStartPlaySoundObject(
		    (YConnection *)recorder->con,
		    object,
		    &value
		);
		if(yid == YIDNULL)
		    return(NULL);

		/* Allocate sound play structure. */
		snd_play = (snd_play_struct *)calloc(
		    1, sizeof(snd_play_struct)
		);
		if(snd_play == NULL)
		    return(NULL);

		/* Record values. */
		snd_play->data = (void *)yid;
                snd_play->volume_left = volume_left;
                snd_play->volume_right = volume_right;
                snd_play->sample_rate = sample_rate;
		snd_play->options = options;
	    }
#endif	/* Y_H */
            break;

        }

	return(snd_play);
}

/*
 *	Same as SoundStartPlay() except no reference is returned and there
 *	is no way to control the sound object being played after it has
 *	started playing (except for calling SoundShutdown()).
 */
void SoundStartPlayVoid(
        snd_recorder_struct *recorder,
        const char *object,	/* Full path to sound object. */
        double volume_left,     /* Volume, from 0.0 to 1.0. */
        double volume_right,
        int sample_rate,        /* Applied sample rate, can be 0. */
        snd_flags_t options     /* Any of SND_PLAY_OPTION_*. */
)
{
        if((recorder == NULL) || (object == NULL))
            return;

        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if(recorder->con != NULL)
            {
                YID yid;
                YEventSoundPlay value;


                /* Set up sound play values. */
                value.flags = YPlayValuesFlagPosition |
                              YPlayValuesFlagTotalRepeats |
                              YPlayValuesFlagVolume |
                              YPlayValuesFlagSampleRate;
                value.position = 0;
                value.total_repeats = ((options & SND_PLAY_OPTION_REPEATING) ?
                    -1 : 1
                );
                value.left_volume = ((options & SND_PLAY_OPTION_MUTE) ?
                    0 : volume_left
                );
                value.right_volume = ((options & SND_PLAY_OPTION_MUTE) ?
                    0 : volume_right
                );
                value.sample_rate = sample_rate;
            
                /* Start playing sound object. */
                yid = YStartPlaySoundObject(
                    (YConnection *)recorder->con,
                    object,
                    &value
                );
                if(yid == YIDNULL)
                    return;
            }
#endif  /* Y_H */
            break;

        }
}

/*
 *	Mutes or unmutes a sound object already playing.
 */
void SoundPlayMute(
        snd_recorder_struct *recorder, snd_play_struct *snd_play,
	int mute
)
{
        if((recorder == NULL) || (snd_play == NULL))
            return;

	if(mute)
	    snd_play->options |= SND_PLAY_OPTION_MUTE;
	else
	    snd_play->options &= ~SND_PLAY_OPTION_MUTE;

	SoundChangePlayVolume(
	    recorder, snd_play,
	    snd_play->volume_left, snd_play->volume_right
	);
}

/*
 *	Changes the volume of the sound object already playing.
 */
void SoundChangePlayVolume(
        snd_recorder_struct *recorder, snd_play_struct *snd_play,
        double volume_left,     /* Volume, from 0.0 to 1.0. */
        double volume_right
)
{
        if((recorder == NULL) || (snd_play == NULL))
            return;

	/* Change in volume? */
	if((volume_left == snd_play->volume_left) &&
           (volume_right == snd_play->volume_right)
	)
	    return;

        switch(recorder->type)   
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if((recorder->con != NULL) &&
               (snd_play->data != NULL)
	    )
            {
                YEventSoundPlay value;


                /* Set up sound play values. */
                value.flags = YPlayValuesFlagVolume;
                value.left_volume = ((snd_play->options & SND_PLAY_OPTION_MUTE) ?
		    0 : volume_left
		);
                value.right_volume = ((snd_play->options & SND_PLAY_OPTION_MUTE) ?
		    0 : volume_right
		);

		/* Change play sound object values. */
                YSetPlaySoundObjectValues(
		    (YConnection *)recorder->con,
		    (YID)snd_play->data,
		    &value
		);

		/* Record values. */
                snd_play->volume_left = volume_left;
                snd_play->volume_right = volume_right;
            }
#endif  /* Y_H */
            break;

        }
}

/*
 *      Changes the sample rate of the sound object already playing.
 */
void SoundChangePlaySampleRate(
        snd_recorder_struct *recorder,
        snd_play_struct *snd_play,
        int sample_rate         /* Applied sample rate, can be 0. */
)
{
        if((recorder == NULL) || (snd_play == NULL))
            return;

        /* Change in applied sample rate? */
        if(sample_rate == snd_play->sample_rate)
            return;

        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if((recorder->con != NULL) &&
               (snd_play->data != NULL)
            )
            {
                YEventSoundPlay value;


		if(sample_rate < 0)
                    sample_rate = 0;

                /* Set up sound play values. */
                value.flags = YPlayValuesFlagSampleRate;
                value.sample_rate = sample_rate;

                /* Change play sound object values. */
                YSetPlaySoundObjectValues(
                    (YConnection *)recorder->con,
                    (YID)snd_play->data,
                    &value
                );

		/* Record values. */
		snd_play->sample_rate = sample_rate;
            }
#endif  /* Y_H */
            break;

        }
}

/*
 *	Stops a sound object already playing and deallocates the given
 *	snd_play structure. The snd_play structure should not be
 *	referenced again after this call. The snd_play structure will be
 *	deallocated even if recorder is NULL. However recorder should 
 *	always be given if available.
 */
void SoundStopPlay(
        snd_recorder_struct *recorder, snd_play_struct *snd_play
)
{
#define DO_FREE_PLAY_STRUCT	\
if(snd_play != NULL) \
{ \
 free(snd_play); \
 snd_play = NULL; \
}

	if(recorder == NULL)
	{
	    DO_FREE_PLAY_STRUCT
	    return;
	}

	if(snd_play == NULL)
	    return;

        switch(recorder->type)
        {       
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if((recorder->con != NULL) &&
               ((YID)snd_play->data != YIDNULL)
            )
            {
		YDestroyPlaySoundObject(
                    (YConnection *)recorder->con,
                    (YID)snd_play->data
		);
		snd_play->data = (void *)YIDNULL;
	    }
#endif	/* Y_H */
	    break;

	}

	DO_FREE_PLAY_STRUCT
#undef DO_FREE_PLAY_STRUCT
}


/*
 *	Starts playing the given background music sound object specified
 *	by object. If an existing background music sound object is being
 *	played then it will be stopped first.
 *
 *	The given repeats can be -1 for infinate repeating (only stopped
 *	when another background music sound object begins to play or when
 *	the sound server shuts down).
 *
 *	Returns non-zero on error.
 */
int SoundMusicStartPlay(
        snd_recorder_struct *recorder,
        const char *object,     /* Full path to object. */
        int repeats             /* Number of repeats, -1 for infinate. */
)
{
	snd_play_struct *snd_play = NULL;


	if((recorder == NULL) || (object == NULL))
	    return(-1);

	/* Stop currently playing background music if any. */
	SoundMusicStopPlay(recorder);

	/* Unable to destroy background music sound object? */
	if(recorder->background_music_sndobj != NULL)
	    return(-1);


        switch(recorder->type)
        {
          case SNDSERV_TYPE_Y:
#ifdef Y_H
            if(recorder->con != NULL)
            {
                YID yid;
                YEventSoundPlay value;


                /* Set up sound play values. */
                value.flags = YPlayValuesFlagPosition |
                              YPlayValuesFlagTotalRepeats |
                              YPlayValuesFlagVolume |
                              YPlayValuesFlagSampleRate;
                value.position = 0;
                value.total_repeats = repeats;
                value.left_volume = 1.0;
                value.right_volume = 1.0;
                value.sample_rate = recorder->sample_rate;

                /* Start playing sound object. */
                yid = YStartPlaySoundObject(
                    (YConnection *)recorder->con,
                    object,
                    &value
                );
                if(yid == YIDNULL)
                    return(-1);

                /* Allocate background music sound play structure. */
		recorder->background_music_sndobj = snd_play = (snd_play_struct *)calloc(
		    1, sizeof(snd_play_struct)
		);
                if(snd_play == NULL)
                    return(-1);

                /* Record values. */
                snd_play->data = (void *)yid;
                snd_play->volume_left = value.left_volume;
                snd_play->volume_right = value.right_volume;
                snd_play->sample_rate = value.sample_rate;
                snd_play->options = 0;
            }
#endif  /* Y_H */
            break;

        }

	return(0);
}

/*
 *	Returns true if there is a background music sound object playing
 *	on the specified recorder.
 */
int SoundMusicIsPlaying(snd_recorder_struct *recorder)
{
        if((recorder != NULL) ?
	    (recorder->background_music_sndobj != NULL) : 0
	)
            return(1);
	else
	    return(0);
}

/*
 *	Stops playing the background music sound object (if any).
 */
void SoundMusicStopPlay(snd_recorder_struct *recorder)
{
	if(recorder == NULL)
	    return;

	if(recorder->background_music_sndobj != NULL)
	{
	    SoundStopPlay(
		recorder,
		recorder->background_music_sndobj
	    );
	    recorder->background_music_sndobj = NULL;
	}
/* printf("SoundMusicStopPlay()\n"); */
}
